package com.me.brickbuster.state;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.physics.box2d.*;
import com.badlogic.gdx.utils.Array;
import com.me.brickbuster.BrickBuster;
import com.me.brickbuster.entity.*;
import com.me.brickbuster.entity.powerup.PowerUp;
import com.me.brickbuster.layout.*;
import com.me.brickbuster.physics.Box2dContactListener;
import com.me.brickbuster.physics.EntityType;

import java.util.Iterator;

public class PlayState extends FieldState {

	public static final float BOARD_WIDTH = 54f;
	public static final float BOARD_HEIGHT = 96f;
	public static final float PIXEL_PER_METER = BrickBuster.BOARD_WIDTH / BOARD_WIDTH;
	public static final float EDGE_PADDING = 1.7f;

	public static final Vector2 lowerLeftCorner =
			new Vector2(EDGE_PADDING,0);
	public static final Vector2 lowerRightCorner =
			new Vector2(BrickBuster.BOARD_WIDTH/PIXEL_PER_METER-EDGE_PADDING,0);
	public static final Vector2 upperRightCorner =
			new Vector2(BrickBuster.BOARD_WIDTH/PIXEL_PER_METER-EDGE_PADDING, BrickBuster.BOARD_HEIGHT/PIXEL_PER_METER-EDGE_PADDING);
	public static final Vector2 upperLeftCorner =
			new Vector2(EDGE_PADDING, BrickBuster.BOARD_HEIGHT/PIXEL_PER_METER-EDGE_PADDING);

	public Body playArea;
	public Array<Body> bodies;

	public LevelLoader levelLoader;
	public Level currentLevel;

	public Array<PowerUp> powerUps;
	public Paddle paddle;
	public Array<Ball> balls;
	public Array<Brick> bricks;
	public Array<Shield> shields;

	private float updateTime = 0f;

	private final Box2DDebugRenderer debugRenderer;

	public PlayState(BrickBuster game) {
		super(game);
		debugRenderer = new Box2DDebugRenderer();
	}

	@Override
	public void setup() {
		// Initialize a world with no gravity
		world = new World(new Vector2(), false);
		world.setContactListener(new Box2dContactListener());
		bodies = new Array<Body>();

		levelLoader = new GridLevelLoader(this);
		currentLevel = levelLoader.getNextLevel();

		initializeLevel();
	}

	@Override
	public void render() {
		world.getBodies(bodies);
		for (Body b : bodies) {
			Entity e = (Entity) b.getUserData();
			if (e instanceof Ball || e instanceof PowerUp || e instanceof Brick) {
				e.setPos(b.getPosition());
			}
		}

		long start = System.nanoTime();
		for (Brick block : bricks) {
			block.render(game.sr);
		}
		for (PowerUp powerUp : powerUps) {
			powerUp.render(game.sr);
		}
		for (Ball ball : balls) {
			ball.render(game.sr);
		}
		for (Shield shield : shields) {
			shield.render(game.sr);
		}
		paddle.render(game.sr);

		//debugRenderer.render(world, game.cam.combined.cpy().scl(PIXEL_PER_METER));
		long renderTime = System.nanoTime() - start;

		game.sb.begin();
		game.font.setColor(Color.GRAY);
		game.font.draw(game.sb, String.format("FPS: %d Update: %.2f ms Render: %.2f ms",
				Gdx.graphics.getFramesPerSecond(), updateTime/1000000f, renderTime/1000000f), 10, BrickBuster.BOARD_HEIGHT-10);
		game.sb.end();
	}

	@Override
	public void update(float dt) {
		long start = System.nanoTime();
		paddle.update(dt);

		for (Iterator<Ball> it = balls.iterator(); it.hasNext();) {
			Ball ball = it.next();
			ball.update(dt);
			if (ball.isDeleted()) {
				it.remove();
				world.destroyBody(ball.getBody());
			}
		}
		if (balls.size == 0) {
			ballReset();
		}

		for (Iterator<PowerUp> it = powerUps.iterator(); it.hasNext();) {
			PowerUp powerUp = it.next();
			powerUp.update(dt);
			if(powerUp.isDeleted()) {
				it.remove();
				world.destroyBody(powerUp.getBody());
			}
		}

		for (Iterator<Shield> it = shields.iterator(); it.hasNext();) {
			Shield shield = it.next();
			shield.update(dt);
			if(shield.isDeleted()) {
				it.remove();
				world.destroyBody(shield.getBody());
			}
		}

		for (Iterator<Brick> it = bricks.iterator(); it.hasNext();) {
			Brick brick = it.next();
			brick.update(dt);
			if (brick.isDeleted()) {
				it.remove();
				brick.dispose();
				world.destroyBody(brick.getBody());
			}
		}
		if (bricks.size == 0) {
			currentLevel = levelLoader.getNextLevel();
			if (currentLevel == null) {
				game.setScreen(new MenuState(game));
				dispose();
			} else {
				initializeLevel();
			}
			return;
		}

		world.step(dt, 6, 2);
		updateTime = System.nanoTime() - start;
	}

	@Override
	public void dispose() {
		super.dispose();
		world.dispose();

		bodies.clear();

		powerUps.clear();
		powerUps = null;
		balls.clear();
		balls = null;
		bricks.clear();
		bricks = null;
		paddle = null;
	}

	private void initializeLevel() {
		world.getBodies(bodies);
		for (Body b : bodies) {
			world.destroyBody(b);
		}

		// Set up the borders
		BodyDef playAreaDef = new BodyDef();
		playAreaDef.type = BodyDef.BodyType.StaticBody;
		playAreaDef.position.set(new Vector2());

		EdgeShape screenEdge = new EdgeShape();
		FixtureDef playAreaFixture = new FixtureDef();
		playAreaFixture.shape = screenEdge;

		playAreaFixture.filter.categoryBits = EntityType.BOUNDARY;
		playAreaFixture.filter.maskBits = EntityType.BALL | EntityType.BRICK;

		playArea = world.createBody(playAreaDef);
		screenEdge.set(lowerRightCorner, upperRightCorner);
		playArea.createFixture(playAreaFixture);
		screenEdge.set(upperRightCorner, upperLeftCorner);
		playArea.createFixture(playAreaFixture);
		screenEdge.set(upperLeftCorner, lowerLeftCorner);
		playArea.createFixture(playAreaFixture);
		screenEdge.dispose();

		powerUps = new Array<PowerUp>();
		paddle = new Paddle(this);

		BrickLayout layout = currentLevel.getLayout();
		layout.initialize();
		bricks = layout.getBricks();

		balls = new Array<Ball>();
		balls.add(new Ball(this));

		shields = new Array<Shield>();
	}

	public int getShieldCount() {
		return shields.size;
	}

	public void addShield() {
		Shield shield = new Shield(this);
		shields.add(shield);
		paddle.setY(paddle.getY() + Shield.SHIELD_HEIGHT);
		for (Ball ball : balls) {
			if (ball.isStuck()) {
				ball.setY(ball.getY() + Shield.SHIELD_HEIGHT);
			}
		}
	}

	public void removeShield(Shield shield) {
		paddle.setY(paddle.getY() - Shield.SHIELD_HEIGHT);
		for (Ball ball : balls) {
			if (ball.isStuck()) {
				ball.setY(ball.getY() - Shield.SHIELD_HEIGHT);
			}
		}
	}

	public void ballReset() {
		Ball ball = new Ball(this);
		ball.setX(paddle.getX());
		ball.setY(Paddle.PADDLE_Y + Paddle.PADDLE_HEIGHT + Ball.RADIUS);
		balls.add(ball);
		paddle.setSticky(false);
		paddle.setWidth(Paddle.DEFAULT_WIDTH);
	}

}
